;;
;; T-Code isearch modification for Mule 2.0.
;;
;; Author : Kaoru Maeda   (maeda@src.ricoh.co.jp)
;; 	    Mikihiko Nakao (nakao)
;;
;; Copyright (C) 1994 Kaoru Maeda and Mikihiko Nakao
;;			

;; $Id: tc-isearch19.el,v 1.9901 1994/11/20 06:31:21 kaoru Exp kaoru $

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License version 1
;; as published by the Free Software Foundation.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;; 94.8.25  Modified for T-Code by M.Nakao
;; 94.10.05 Modified by M.Nakao
;;	* bushu in TC-I-search.
;; 94.11.08 Modified by Kaoru Maeda.
;;	* isearch-delete-char can delete the last incomplete tcode stroke.
;;	* add code for tcode-isearch-start-state.
;;
;; some of following codes are stolen from isearch.el for mule-2.0:
;;
;; isearch.el --- incremental search minor mode.
;; Copyright (C) 1992, 1993, 1994 Free Software Foundation, Inc.
;; Author: Daniel LaLiberte <liberte@cs.uiuc.edu>
;; |\044Date: 1994/05/14 09:50:26 \044|\044Revision: 1.69 \044
;;
;; Comment from Kaoru: strange \044 above denotes Daniel's original RCS Id.

(if (<= (string-to-int emacs-version) 18)
	(error "tc-isearch19 cannot run on Nemacs/mule-1.x.  Use Mule-2.0 or later!"))

;;
;;  Default Binding
;;
(define-key isearch-mode-map "\C-\\" 'isearch-toggle-tcode)

;;
;;  User Variables
;;
(defvar tcode-isearch-start-state nil
  "*T-Code$B%$%s%/%j%a%s%?%k%5!<%A3+;O;~$N(BT-Code$B%b!<%I$r;XDj$9$k!#(B
	nil: $B%P%C%U%!$N(Btcode-on$B$KF14|(B($B%G%U%)%k%H(B)$B!#(B
	t:   $B%P%C%U%!$N(Btcode-on$B$HFHN)!#3+;O;~$O%P%C%U%!$HF1$8!#(B
	0:   $B%P%C%U%!$HFHN)$K>o$K(Bnon-T-Code$B%5!<%A$+$i3+;O!#(B
	1:   $B%P%C%U%!$HFHN)$K>o$K(BT-Code$B%5!<%A$+$i3+;O!#(B
$B%P%C%U%!%m!<%+%kJQ?t!#(B")
(make-variable-buffer-local 'tcode-isearch-start-state)
(setq-default tcode-isearch-start-state nil)

;;
;;  Internal Variables
;; 
(defvar tcode-isearch-active nil
  "$B%$%s%/%j%a%s%?%k%5!<%A$,(B T-Code $B%b!<%I$N$H$-(B non-nil$B!#(B")
(defvar tcode-isearch-first-stroke nil
  "non-nil $B$N>l9g!"(BT-Code $B%$%s%/%j%a%s%?%k%5!<%A$NBh(B1$B%9%H%m!<%/%3!<%I$rJ];}!#(B")
(defvar tcode-isearch-first-char nil
  "non-nil $B$N>l9g!"(BT-Code $B%$%s%/%j%a%s%?%k%5!<%A$NBh(B1$B%9%H%m!<%/$N;zLL$rJ];}!#(B")


(defun isearch-message-prefix (&optional c-q-hack ellipsis nonincremental)
  ;; If about to search, and previous search regexp was invalid,
  ;; check that it still is.  If it is valid now,
  ;; let the message we display while searching say that it is valid.
  (and isearch-invalid-regexp ellipsis
       (condition-case ()
	   (progn (re-search-forward isearch-string (point) t)
		  (setq isearch-invalid-regexp nil
			isearch-within-brackets nil))
	 (error nil)))
  ;; If currently failing, display no ellipsis.
  (or isearch-success (setq ellipsis nil))
  (let ((m (concat (if isearch-success "" "failing ")
		   (if isearch-wrapped "wrapped ")
		   (if isearch-word "word " "")
		   (if isearch-regexp "regexp " "")
		   (if tcode-isearch-active "TC-" "")
		   (if nonincremental "search" "I-search")
		   (if isearch-forward ": " " backward: ")
		   )))
    (aset m 0 (upcase (aref m 0)))
    m))


(defun isearch-toggle-tcode ()
  "$B%$%s%/%j%a%s%?%k%5!<%ACf$N(B T-Code $B%b!<%I$r%H%0%k$9$k!#(B
$B%$%s%/%j%a%s%?%k%5!<%ACf$K(B prefix/postfix $BIt<sJQ49$,;H$($^$9!#(B"
  (interactive)
  (setq tcode-isearch-active (not tcode-isearch-active))
  (setq tcode-isearch-first-stroke nil
	tcode-isearch-first-char nil)
  (if tcode-isearch-start-state
      ()
    (toggle-tcode-mode))
  (isearch-update))

(defun isearch-printing-char ()
  "Add this ordinary printing character to the search string and search."
  (interactive)
  (if (not tcode-isearch-active)
	  (isearch-process-search-char (isearch-last-command-char))
	(let ((c (isearch-last-command-char)) k)
	  (setq k (tcode-get-key-address c))
	  (if (not tcode-isearch-first-stroke)
	      (if (< k 0)
		  (isearch-process-search-char c)
		(setq tcode-isearch-first-stroke k
		      tcode-isearch-first-char c)
		(isearch-update))
	    (cond
	     ((and (= tcode-isearch-first-stroke 6) (= k 6))   ;; postfix
	      (tcode-isearch-postfix-bushu))
	     ((and (= tcode-isearch-first-stroke 26) (= k 23)) ;; prefix
	      (tcode-isearch-start-bushu))
	     (t
	      (let ((s (tcode-encode-key tcode-isearch-first-stroke k
					 tcode-isearch-first-char   c))
		    (prev (tcode-isearch-bushu)))
		(if (null prev)
		    (isearch-process-search-string s s)
		  (if (null (stringp prev))
		      (isearch-process-search-string "" s)
		    (if (setq s (tcode-lookup-outset prev s))
			(progn
			  (tcode-isearch-end-bushu)
			  (isearch-process-search-string s s))
		      (beep)
		      (isearch-update)))))))
	    (setq tcode-isearch-first-stroke nil
		  tcode-isearch-first-char nil)))))


(defun isearch-delete-char ()
  "Discard last input item and move point back.  
If no previous match was done, just beep."
  (interactive)
  (if (and tcode-isearch-active
	   tcode-isearch-first-stroke)
      (setq tcode-isearch-first-stroke nil
	    tcode-isearch-first-char nil)
    (if (and tcode-isearch-active
	     (tcode-isearch-bushu))
	(setq isearch-message
	      (substring isearch-message
			 0 (string-match ".$" isearch-message)))
      (if (null (cdr isearch-cmds))
	  (ding)
	(isearch-pop-state))))
  (isearch-update))


(defun tcode-isearch-start-bushu ()
  "T-Code $B%$%s%/%j%a%s%?%k%5!<%ACf$N(B prefix $BIt<sJQ49$r;O$a$k!#(B"
  (tcode-bushu-init 2)
  (or tcode-bushu-hash (error "Bushu dictionary not ready."))
  (setq isearch-message (concat isearch-message "$B"%(B"))
  (isearch-update))


(defun tcode-isearch-end-bushu ()
  "T-Code $B%$%s%/%j%a%s%?%k%5!<%ACf$N(B prefix $BIt<sJQ49$r=*$($k!#(B"
  (setq isearch-message
	(substring isearch-message
		   0 (string-match "$B"%(B.$" isearch-message)))
  (isearch-update))


(defun tcode-isearch-postfix-bushu ()
  "T-Code $B%$%s%/%j%a%s%?%k%5!<%ACf$N(B  postfix $BIt<sJQ49!#(B"
  (let ((p1 (string-match "..$" isearch-message))
	(p2 (string-match ".$"  isearch-message)))
    (if (null p1)
	(beep)
      (let ((k1 (substring isearch-message p1 p2))
	    (k2 (substring isearch-message p2)))
	(if (setq k1 (tcode-lookup-outset k1 k2))
	    (progn
	      (isearch-delete-char)
	      (isearch-delete-char)
	      (isearch-process-search-string k1 k1))
	  (beep))))))
    

(defun tcode-isearch-bushu ()
  "isearch-message$BCf$NIt<s9g@.$NJ8;z$rD4$Y$k(B"
  (let ((len (length isearch-message)))
    (cond
     ((string-match "$B"%(B$" isearch-message)
      t)
     ((string-match "$B"%(B.$" isearch-message)
      (substring isearch-message (string-match ".$" isearch-message)))
     (t
      nil))))


(defun tcode-isearch-init ()
  "T-Code $B%$%s%/%j%a%s%?%k%5!<%A$N=i4|2=!#(B"
  (setq tcode-isearch-first-stroke nil
	tcode-isearch-first-char nil
	tcode-isearch-active (if (numberp tcode-isearch-start-state)
			     (if (zerop tcode-isearch-start-state) nil t)
			   tcode-on))
  (isearch-update))

(add-hook 'isearch-mode-hook 'tcode-isearch-init)
