;;
;; T-Code frontend for Mule 2.0.
;;
;; Author : Kaoru Maeda   (maeda@src.ricoh.co.jp)
;;	    Yasushi Saito (yasushi@is.s.u-tokyo.ac.jp)
;;
;; Copyright (C) 1989--1994 Kaoru Maeda and Yasushi Saito.
;;

;; $Id: tc-mule2.el,v 1.9901 1994/11/20 06:28:59 kaoru Exp kaoru $

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License version 1
;; as published by the Free Software Foundation.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(or (boundp 'MULE)
    (error "tc-mule2 cannot run on Nemacs nor Emacs.  Use Mule!"))
(if (<= (string-to-int emacs-version) 18)
    (error "tc-mule2 cannot run under Emacs-18 based Mule.  Use Mule-2.0 or later!"))

(provide 'tc-mule2)

;;
;; Version
;;

(defun tcode-version ()
  "T-Code $B%I%i%$%P$N%P!<%8%g%s$rI=<($9$k!#(B"
  (interactive)
  (if (interactive-p)
      (message "T-Code driver for Mule version %s" (tcode-version))
    (substring "$Revision: 1.9901 $" 11 -2)))

;;
;; Variables for customization
;;
(defvar tcode-use-egg-in-mazegaki nil
  "* $B8r$<=q$-JQ49$r$?$^$4$G9T$&;~$K$O(Bt$B!"<+A0$G9T$&;~$K$O(Bnil$B$K$9$k!#(B")

(or (featurep 'egg)
    ;; override user preference
    (setq tcode-use-egg-in-mazegaki nil))

(defvar tcode-self-inserting-commands
  (nconc (if (featurep 'egg) '((fence-self-insert-command)
			       (egg-self-insert-command . "p")))
	 '((electric-c-brace . "P")
	   (electric-c-semi . "P")
	   (electric-c-terminator . "P")
	   (electric-perl-terminator . "P")))
  "*tcode-on $B$N>l9g$K$O(B tcode-self-insert-command $B$K2=$1$F$[$7$$%3%^%s%I$N(Balist$B!#(B
(command . arg) $B$N7A$r$7$F$$$k!#(Barg $B$O(B interactive $B$NBhFsJQ?t$HF1$8J8;zNs!#(B")

(defvar tcode-alnum-1-to-2-table
  "$B!!!*!I!t!p!s!u!G!J!K!v!\!$!]!%!?#0#1#2#3#4#5#6#7#8#9!'!(!c!a!d!)!w#A#B#C#D#E#F#G#H#I#J#K#L#M#N#O#P#Q#R#S#T#U#V#W#X#Y#Z!N!o!O!0!2!F#a#b#c#d#e#f#g#h#i#j#k#l#m#n#o#p#q#r#s#t#u#v#w#x#y#z!P!C!Q!1(B"
  "* STRING: 1$B%P%$%H1Q?t;z(B ' '..'~' $B$r(B2$B%P%$%H1Q?t;z$XJQ49(B/$B5UJQ49$9$k$?$a$N%F!<%V%k(B")

(defvar tcode-alnum-2-byte-p nil
  "* $B1Q?t;z$N%P%$%HD9@Z$j49$(%U%i%0!#(Bt $B$G$O(B2$B%P%$%H7O!"(Bnil $B$G(B1$B%P%$%H7O(B")
(make-variable-buffer-local 'tcode-alnum-2-byte-p)

(defvar tcode-kutouten t
  "* $B6gFIE@@Z$j49$(%U%i%0!#(Bt:$B!V!"!#!W(B; nil:$B!V(B, . $B!W(B; $B$^$?$O(B (STR . STR) $B$N7A$N(B alist$B!#(B
nil $B$O(B '((\"$B!"(B\" . \", \") (\"$B!#(B\" . \". \"))$B$HEy2A!#(B")
(make-variable-buffer-local 'tcode-kutouten)

(defvar tcode-dictionary-name "~/emacs/lisp/mazegaki.dic"
  "* $B8r$<=q$-JQ49<-=q$N%Q%9L>(B")
(defvar tcode-bushu-dictionary-name "~/emacs/lisp/bushu-mule.dic"
  "* $BIt<sJQ49<-=q$N%Q%9L>(B")
(defvar tcode-stroke-file-name "~/emacs/lisp/tcode.st"
  "* $B%9%H%m!<%/I=$N%Q%9L>(B")
(defvar tcode-table-file-name "~/emacs/lisp/tc-mule.tbl.el"
  "* Tcode$B%F!<%V%kI=$N%Q%9L>(B")
(defvar tcode-record-file-name "~/.tc-record"
  "* non-nil $B$N$H$-!"(BT-Code$B$NE}7W$r$3$N%U%!%$%k$K5-O?$9$k(B")

(defvar tcode-mazegaki-gakushu-level 2
  "* $B8r$<=q$-JQ49$N3X=,%l%Y%k(B
	0 : $B3X=,$7$J$$(B
	1 : $BF1$8FI$_$N=O8l$NCf$G$O:G6a$K;H$C$?=O8l$r@h$K=P$9$h$&$K$9$k(B
	2 : $B<-=q$NEPO?=g=x$b=q$-$+$($F!":G6a$K;H$C$?=O8l$rAa$/(B
		$B=P$9$3$H$,$G$-$k$h$&$K$9$k!#(B")

(defvar tcode-kakutei-register ?\[
  "* $B8r$<=q$-JQ49$G:G8e$K3NDj$7$?J8;zNs$rJ]B8$7$F$*$/%l%8%9%?!#(B
nil $B$N>l9g$K$OJ]B8$5$l$J$$!#(B")

(defvar tcode-bushu-on-demand 1
  "* $BIt<sJQ49<-=q$r$$$D=i4|2=$9$k$+!#(B
	0 : tc.el $B%m!<%I;~(B
	1 : $B=i$a$F(BT-Code$B%b!<%I$KF~$C$?$H$-(B
	2 : prefix$BIt<sJQ49$r3+;O$7$?$H$-!"$^$?$OJ8;z$N%X%k%W$r8+$h$&$H$7$?$H$-(B
	3 : postfix$BIt<sJQ49$r3+;O$7$?$H$-(B")

(defvar tcode-special-commands-alist
  '((0 0 (tcode-show-tables nil nil))	; 11 : LL$BI=$NI=<((B
    (0 9 (tcode-show-tables nil t))	; 10 : LR$BI=$NI=<((B
    (9 0 (tcode-show-tables t nil))	; 01 : RL$BI=$NI=<((B
    (9 9 (tcode-show-tables t t))	; 00 : RR$BI=$NI=<((B
    (1 1 (tcode-start-jiscode))		; 22 : JIS $B%3!<%II=F~NO(B
    (2 2 (progn (setq tcode-alnum-2-byte-p (not tcode-alnum-2-byte-p)) ; 33 : 1-2$B%P%$%H@Z$j49$((B
		(tcode-mode-line-redisplay)))
    (2 1 (cond ((consp tcode-kutouten))	; 32 : $B6gFIE@$N%H%0%k(B
	       (t (setq tcode-kutouten (not tcode-kutouten)))))
    (8 9 (progn				; 90 : $BA43Q%b!<%I(B
	   (setq tcode-zenkaku-mode (not tcode-zenkaku-mode))
	   (tcode-mode-line-redisplay)))
    (3 3 (tcode-display-stroke-for-char tcode-help-char)) ; 44 : $B%X%k%W(B
    (4 4 (tcode-query-stroke (point)))	; 55 : $B%X%k%W(B
    (6 6 (tcode-compose-chars))		; 77 : postfix $BIt<sJQ49(B
    (7 7 (tcode-transpose-strokes nil))	; 88 : transpose-strokes
    (8 8 (tcode-clear))			; 99 : $BIt<sJQ49$N%-%c%s%;%k(B
    (26 23 (tcode-start-bushu))		; jf : $BIt<sJQ49$N3+;O(B
    (23 26 (if tcode-use-egg-in-mazegaki ; fj : $B8r$<=q$-JQ49$N3+;O(B
	       (progn
		 (setq egg:*mode-on* t
		       egg:*input-mode* nil)
		 (setq tcode-in-mazegaki-p t)
		 (toggle-egg-mode))
	     (tcode-start-mazegaki))))
  "*tcode-on $B$N$H$-!"$3$l$i$N%9%H%m!<%/$OJ8;zF~NO$G$J$/%3%^%s%I$r<B9T$9$k(B")

(defvar tcode-mode-map nil
  "*tcode-on $B$N$H$-(B T-Code $B%-!<0J30$N%-!<$N$?$a$N%-!<%^%C%W(B")

(if tcode-mode-map ()
  (setq tcode-mode-map (make-sparse-keymap)))

(defvar tcode-ready-hook nil
  "* Mule$B$rN)$A>e$2$F$+$i:G=i$K(Btcode-mode$B$KF~$C$?$H$-<B9T$5$l$k(B")
(defvar tcode-bushu-ready-hook nil
  "* $BIt<sJQ49$N=i4|2=D>8e$K<B9T$5$l$k(B")
(defvar tcode-mode-hook nil
  "* $B$=$N%P%C%U%!$G=i$a$F(Btcode-mode$B$KF~$C$?$H$-<B9T$5$l$k(B")
(defvar tcode-toggle-hook nil
  "* tcode-mode$B$r%H%0%k$9$kEY$K<B9T$5$l$k(B")

(defvar tcode-help-first-stroke "$B!{(B"
  "* $B%X%k%WI=$GBh0lBG800LCV$rI=$o$9J8;z(B")
(defvar tcode-help-second-stroke "$B!|(B"
  "* $B%X%k%WI=$GBhFsBG800LCV$rI=$o$9J8;z(B")
(defvar tcode-help-same-stroke "$B!}(B"
  "* $B%X%k%WI=$GBh0lBG80!"BhFsBG80$,$H$b$K$3$N0LCV$G$"$k$3$H$rI=$o$9J8;z(B")

(defvar tcode-auto-zap-table nil
  "* non-nil $B$N$H$-(B 1 0 $B$J$I$GI=<($5$l$k%9%H%m!<%/I=$r<!$NBG80$G<+F0E*$K>C$9(B")

(defvar tcode-keymap-table
;;   0  1  2  3  4  5  6  7  8  9 
;;  10 11 12 13 14 15 16 17 18 19
;;  20 21 22 23 24 25 26 27 28 29
;;  30 31 32 33 34 35 36 37 38 39

;;   1  2  3  4  5  6  7  8  9  0
;;   q  w  e  r  t  y  u  i  o  p
;;   a  s  d  f  g  h  j  k  l  ;
;;   z  x  c  v  b  n  m  ,  .  /

;;      !  \"   #   $   %   &   '   (   )   *   +   ,   -   .   / 
;;  0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?
;;  @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
;;  P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _
;;  `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o
;;  p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~
  [
   -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  37  -1  38  39
   09  00  01  02  03  04  05  06  07  08  -1  29  -1  -1  -1  -1
   -1  -2  -2  -2  -2  -2  -2  -2  -2  -2  -2  -2  -2  -2  -2  -2
   -2  -2  -2  -2  -2  -2  -2  -2  -2  -2  -2  -1  -1  -1  -1  -1
   -1  20  34  32  22  12  23  24  25  17  26  27  28  36  35  18
   19  10  13  21  14  16  33  11  31  15  30  -1  -1  -1  -1
   ]
  "* T-Code $B%-!<%F!<%V%kI=!#$=$N0lJ8;z$rF~NO$7$?$H$-$N0UL#$rI=$o$9!#(B
0..39:	T-Code $B%-!<!#(B
-1:	$B$=$NJ8;z!#(B
-2:	$BBP1~$9$k1Q>.J8;z!#(B
-3:	tcode-mode-map $B$K$7$?$,$C$?%3%^%s%I!#(B
< -3:	- ($BJ8;z%3!<%I(B)$B!#(B")

;;
;; Global Variables
;;
(defvar tcode-tbl nil "T-Code$B$N%9%H%m!<%/I=!#(Bchar$B$NG[Ns$NG[Ns$K$J$C$F$$$k!#(B")
(defvar tcode-help-char nil "$B$X%k%W$NBP>]J8;z(B")
(defvar mazegaki-buffer nil "$B8r$<=q$-JQ49<-=qMQ%P%C%U%!(B")
(defvar bushu-buffer nil "$BIt<sJQ49<-=qMQ%P%C%U%!(B")
(defvar tcode-bushu-hash nil
  "$BIt<sJQ49%O%C%7%eI=!#<BBN$O(B Elisp $B$N(B obarray$B!#(B")
(defvar tcode-bushu-reverse-hash nil
  "$BIt<sJQ495U$S$-%O%C%7%eI=!#<BBN$O(B Elisp $B$N(B obarray$B!#(B
$BJ8;z$+$i%9%H%m!<%/$r0z$/$?$a$NI=$b7s$M$F$$$k(B('stroke $B%W%m%Q%F%#(B)$B!#(B")

(defvar tcode-number-strokes 0)
(defvar tcode-self-insert-non-undo-count 0)
(defvar tcode-input-chars 0)
(defvar tcode-bushu-occurrence 0)
(defvar tcode-mazegaki-occurrence 0)
(defvar tcode-special-occurrence 0)

(defvar tcode-on-in-minibuffer nil "global variable")

;;
;; Buffer Local Variables
;;
(defvar tcode-on nil "T-Code $B%b!<%I$N$H$-(Bt$B!#(B")
(make-variable-buffer-local 'tcode-on)
(setq-default tcode-on nil)

(defvar tcode-zenkaku-mode nil "$BA43Q%b!<%I$N$H$-(Bt$B!#(Btcode-on $B$,(B t $B$N$H$-$N$_M-8z!#(B")

(defvar tcode-radicals-max 1)
(defvar tcode-bushu-nest 0 "$BIt<sJQ49$N%M%9%H%l%Y%k(B")
(make-variable-buffer-local 'tcode-bushu-nest)
(defvar tcode-in-mazegaki-p nil "$B8r$<=q$-JQ49%b!<%I$+(B?")
(make-variable-buffer-local 'tcode-in-mazegaki-p)
(defvar tcode-ready-in-this-buffer nil "$B$3$N%P%C%U%!Fb$G(BT-Code$B$N=`Hw$,(BOK")
(make-variable-buffer-local 'tcode-ready-in-this-buffer)
(setq-default tcode-ready-in-this-buffer nil)


;;
;; T-Code main driver
;;

(global-set-key "\C-\\" 'toggle-tcode-mode)

(defun toggle-tcode-mode ()
  "T-Code$B%b!<%I$r%H%0%k$9$k!#(B
T-Code$B%b!<%ICf$N%-!<A`:n$O<!$NDL$j!#(B
   jf : $BIt<sJQ49%b!<%I$KF~$k!#(Bjf$B$rBG$AB3$1$k$H:F5"E*$KIt<sJQ49$r(B
	$B9T$&$3$H$,$G$-$k!#(B
   fj : $B8r$<=q$-JQ49%b!<%I$KF~$k!#(B
          $B$3$N%b!<%I$G(B' '$B$r$*$9$H8r$<=q$-JQ49$r9T$&!#(B
	  $BJQ49$r9T$C$F$$$k:GCf$O<!$N%-!<$,M-8z$G$"$k!#(B
		^P : $BA08uJd(B
		^N : $B<!8uJd(B
		^T : $BEPO?(B
   00, 01, 10, 11 : T-Code$B$NI=$rI=<($9$k!#(B
   22 : JIS $B%3!<%I0lMwI=$K$h$kF~NO!#(B
   32 : $B!"!#$H(B, . $B$r@Z$j$+$($k!#(B
   33 : T-Code$BI=$K$"$k5-9f$NA4H>3Q@Z$j49$(!#(B
   44 : $BD>A0$KF~NO$7$?J8;z$NBG$AJ}$rI=<($9$k!#(B
   55 : $B%]%$%s%H0LCV$K$"$kJ8;z$NBG$AJ}$rI=<($9$k!#(B
   77 : $B%]%$%s%HA0$K$"$kFsJ8;z$GIt<sJQ49$r9T$J$&!#(B
   88 : $B%]%$%s%H0LCV$K$"$kJ8;z$r5U%9%H%m!<%/2=$9$k(B($BNc(B: $BL#(B->$B$N(B)$B!#(B
        $B9TKv$G$O%]%$%s%H$ND>A0$NJ8;z$rJQ49$9$k!#(B
   90 : $BA43QF~NO%b!<%I!#(BT-Code$B%b!<%I$rH4$1$k$^$GM-8z!#(B
   99 : $B8r$<=q$-JQ49%b!<%I$dIt<sJQ49%b!<%I$K$$$?;~$K!"(B
	$B$=$l$i$rA4It%-%c%s%;%k$9$k(B
   ^\\ : T-Code$B%b!<%I$rH4$1$k(B

Mule$B$r5/F0$7$F0JMh=i$a$F<B9T$5$l$k$H(B tcode-init $B$r!"(B
$B8=%P%C%U%!$G=i$a$F<B9T$5$l$k$H(B tcode-buffer-init $B$r$3$N=g$K<B9T$9$k!#(B
tcode-on $B$r?7$7$$CM$K@_Dj$7$?$N$A(B tcode-toggle-hook $B$r<B9T$9$k!#(B
"
  (interactive)
  (or (featurep 'tcode-ready) (tcode-init))
  (or tcode-ready-in-this-buffer (tcode-buffer-init))
  (if (string-match " \\*Mini" (buffer-name (current-buffer)))
      (setq tcode-on-in-minibuffer (not tcode-on-in-minibuffer)
	    tcode-on tcode-on-in-minibuffer)
    (setq tcode-on (not tcode-on)))
  (or (tcode-on)
      (setq tcode-zenkaku-mode nil))
  (tcode-mode-line-redisplay)
  (run-hooks 'tcode-toggle-hook)
  (tcode-clear)
  (setq tcode-self-insert-non-undo-count 1)
  (if tcode-in-mazegaki-p
      (tcode-mazegaki-kakutei)))

(defun tcode-init ()
  "Mule$B$,5/F0$5$l$F$+$i:G=i$N(Btoggle-tcode-mode$B$G<B9T$5$l$k!#(B
$B%3!<%II=!"8r$<=q$-JQ49$N=i4|2=$r9T$J$&!#(B
tcode-ready-hook $B$r8F$V!#(B"
  (or (featurep 'tcode-ready)
      (progn
	(tcode-load-table tcode-table-file-name)
	(tcode-bushu-init 1)
	(if (featurep 'egg)
	    ()
	  (substitute-key-definition 'self-insert-command
				     'tcode-self-insert-command-maybe
				     global-map))
	(provide 'tcode-ready)
	(run-hooks 'tcode-ready-hook))))

(defun tcode-buffer-init ()
  "$B%P%C%U%!$4$H$N(BT-Code$B$N=i4|2=$r9T$J$&!#(Btcode-mode-hook$B$r8F$V!#(B"
  (or tcode-ready-in-this-buffer
      (progn
	(tcode-substitute-self-insertings)
	(setq tcode-ready-in-this-buffer t)
	(run-hooks 'tcode-mode-hook))))

(defun tcode-on ()
  "Return tcode-on or tcode-on-in-minibuffer according to current buffer."
  (if (string-match " \\*Mini" (buffer-name (current-buffer)))
      tcode-on-in-minibuffer
    tcode-on))

;; mode line
(defconst tcode-transparent-mode-indicator   "--")
(defconst tcode-zenkaku-mode-indicator       "$B#A(B")
(defconst tcode-tcode-mode-indicator         "TC")
(defconst tcode-zenkaku-tcode-mode-indicator "$B#T(B")

(defvar tcode-mode-indicator "--")
(make-variable-buffer-local 'tcode-mode-indicator)

(defconst tcode-egg-mode-line
  (and (featurep 'egg)
       (fboundp 'mode-line-egg-mode-update)))

(defun tcode-mode-line-redisplay ()
  (setq tcode-mode-indicator
	(cond ((tcode-on)
	       (cond (tcode-zenkaku-mode tcode-zenkaku-mode-indicator)
		     (tcode-alnum-2-byte-p tcode-zenkaku-tcode-mode-indicator)
		     (t tcode-tcode-mode-indicator)))
	      (t tcode-transparent-mode-indicator)))
  (if tcode-egg-mode-line
      (mode-line-egg-mode-update tcode-mode-indicator)
    (and (string-match " \\*Mini" (buffer-name (current-buffer)))
	 (boundp 'minibuffer-preprompt)
	 (setq minibuffer-preprompt (list "[" tcode-mode-indicator "]")))
    (set-buffer-modified-p (buffer-modified-p))))

(defun tcode-find-symbol-in-tree (item tree)
  (if (consp tree)
      (or (tcode-find-symbol-in-tree item (car tree))
	  (tcode-find-symbol-in-tree item (cdr tree)))
    (equal item tree)))

(if tcode-egg-mode-line ()
  (or (tcode-find-symbol-in-tree 'tcode-mode-indicator
				 (default-value 'mode-line-format))
      (setq-default mode-line-format
		    (cons '(mc-flag ("[" tcode-mode-indicator "]"))
			  (default-value 'mode-line-format)))))

(defun tcode-substitute-self-insertings ()
  (let ((commands tcode-self-inserting-commands)
	com
	symbol
	havearg
	tcode-func
	orig-func)
    (while commands
      (setq com (car commands)
	    commands (cdr commands)
	    symbol (car com)
	    havearg (cdr com)
	    tcode-func (intern (concat "tcode:" (symbol-name symbol)))
	    orig-func (intern (concat "orig:" (symbol-name symbol))))
      (and (not (fboundp tcode-func))
	   (fboundp symbol)
	   (progn
	     (fset orig-func (symbol-function symbol))
	     (eval (list 'defun tcode-func (if havearg '(arg))
			 (documentation orig-func t)
			 (cons 'interactive (if havearg (list havearg)))
			 (list 'if 'tcode-on '(tcode-self-insert-command)
			       (cons orig-func (if (cdr com) '(arg))))))
	     (fset symbol tcode-func))))))

(defun tcode-self-insert-command-maybe (arg)
  (interactive "p")
  (if (tcode-on)
      (tcode-self-insert-command arg)
    (self-insert-command arg)))

(defun tcode-exit-minibuffer ()
  (interactive)
  (setq tcode-on nil
	tcode-zenkaku-mode nil
	tcode-on-in-minibuffer nil))

(add-hook 'minibuffer-exit-hook 'tcode-exit-minibuffer)

(defun tcode-clear ()
  "$B$d$d$3$7$$%b!<%I$KF~$C$F$$$k$N$rA4It%/%j%"$9$k(B"
  (interactive)
  (if (tcode-on)
      (progn
	(while (< 0 tcode-bushu-nest) (tcode-end-bushu))
	(if tcode-in-mazegaki-p
	    (tcode-mazegaki-kakutei)))))

(defconst tcode-null-map (make-keymap))

(defvar tcode-self-insert-non-undo-count 0)

(defun tcode-self-insert-command (&optional arg)
  "Encode Tcode character and insert."
  (interactive "*")
  (if (or
       (not (eq last-command this-command))
       (>= tcode-self-insert-non-undo-count 20))
      (setq tcode-self-insert-non-undo-count 1)
    (cancel-undo-boundary)
    (setq tcode-self-insert-non-undo-count
	  (1+ tcode-self-insert-non-undo-count)))
  (if (and tcode-use-egg-in-mazegaki
	   (not buffer-read-only)
	   mc-flag
	   egg:*mode-on* egg:*input-mode*
	   (not (eq (current-local-map) fence-mode-map))
	   (not egg:*in-fence-mode*) ;;; inhibit recursive fence mode
	   (not (= last-command-char  ?  )))
      (egg:enter-fence-mode-and-self-insert)
    (if tcode-zenkaku-mode
	(tcode-insert (tcode-1-to-2 (char-to-string last-command-char)))
      (setq tcode-number-strokes (1+ tcode-number-strokes))
      (let ((k1 (tcode-get-key-address last-command-char))
	    (c1 last-command-char)
	    k2 c2)
	(cond ((= k1 -1)
	       (tcode-insert last-command-char))
	      ((= k1 -2)
	       (tcode-insert (downcase c1)))
	      ((= k1 -3)
	       (let ((save-local-map (current-local-map))
		     (save-global-map (current-global-map))
		     keyseq command)
		 (setq unread-command-events
		       (cons last-command-event unread-command-events))
		 (unwind-protect
		     (progn
		       (use-local-map tcode-mode-map)
		       (use-global-map tcode-null-map)
		       (setq keyseq (read-key-sequence nil))
		       (setq command (key-binding keyseq)))
		   (use-local-map save-local-map)
		   (use-global-map save-global-map))
		 (if command
		     (progn
		       (setq prefix-arg current-prefix-arg)
		       (command-execute command))
		   (ding))))
	      ((< k1 0)
	       (tcode-insert (- k1)))
	      (t
	       (setq c2 (read-char))
	       (setq k2 (tcode-get-key-address c2))
	       (if (< k2 0)
		   ;;2$BBG80L\$,(BT-Code$B%-!<$8$c$J$$!#(B
		   ;; 2$BBG80L\$,(B backward-delete-char $B$@$C$?$i(B1$BBG80L\$r$H$j>C$7(B
		   (let ((command (lookup-key (current-local-map)
					      (make-vector 1 c2))))
		     (cond ((or (eq command 'backward-delete-char)
				(eq command 'backward-delete-char-untabify)))
			   (t
			    (tcode-insert c1)
			    (if (= c2 ?\ ) ()
			      (setq unread-command-events
				    (cons c2 unread-command-events))))))
		 ;;$B@5$7$$(BTcode$BJ8;z$,F~NO$5$l$?(B
		 (let ((special tcode-special-commands-alist)
		       (not-special t)
		       sp)
		   (while special
		     (setq sp (car special)
			   special (cdr special))
		     (and (= (nth 0 sp) k1)
			  (= (nth 1 sp) k2)
			  (progn (setq not-special nil
				       special nil)
				 (setq tcode-special-occurrence (1+ tcode-special-occurrence))
				 (eval (nth 2 sp)))))
		   (if not-special
		       ;; Else, insert the code.
		       (tcode-insert (tcode-encode-key k1 k2 c1 c2)))))))))))

(defvar tcode-bushu-henkan-last-point nil
  "$BIt<sJQ49$r3+;O$7$?;~$N%]%$%s%H0LCV(B")

(defun tcode-insert (ch)
  "CH$B$r%P%C%U%!$KA^F~$9$k!#It<sJQ49Cf$K8F$P$l$k$H(B
$BE,59It<sJQ49$r9T$&!#(B"
  (if (integerp ch) (setq ch (char-to-string ch)))
  (cond
   ((> tcode-bushu-nest 0)
    (let* ((p (point))
	   (p1 (and (>= p 3) (save-excursion (forward-char -1) (point))))
	   (p2 (and (>= p1 2) (save-excursion (goto-char p1) (forward-char -1) (point))))
	   kanji
	   (prev-char (and p1 (buffer-substring p1 p)))
	   (prev-prev-char (and p1 p2 (buffer-substring p2 p1))))
      (if (and prev-char prev-prev-char
	       (string-equal prev-prev-char "$B"%(B")
	       (not (string-equal prev-char "$B"%(B")))
	  (progn
	    ;; $BIt<sJQ49$7$FF~NO(B
	    (or tcode-bushu-henkan-last-point
		(setq tcode-bushu-henkan-last-point (point-marker)))
	    (if (setq kanji (tcode-lookup-outset prev-char ch))
		(progn
		  (delete-region p2 p)
		  (setq tcode-bushu-nest (1- tcode-bushu-nest))
		  (tcode-insert kanji))
	      (beep)
	      (message "")
	      ))
	(setq tcode-bushu-occurrence (1+ tcode-bushu-occurrence))
	(setq tcode-help-char ch)
	;; $BIt<sJQ49$N:`NA$rF~NO(B
	(insert ch)
	)))
   (t
    (setq tcode-input-chars (1+ tcode-input-chars))
    (setq tcode-help-char ch)
    (let ((p (point)))
      (if (and tcode-alnum-2-byte-p
	       (= 1 (length ch)))
	  (setq ch (tcode-1-to-2 ch)))
      (if (= 1 (length ch))
	  ;; ascii iso8859-1$B$N>l9g(B
	  (if tcode-in-mazegaki-p
	      (insert ch)
	    (let ((last-command-char (aref ch 0)))
	      (self-insert-command 1)))
	(insert ch)
	(if tcode-in-mazegaki-p
	    ()
	  (and (boundp 'egg-insert-after-hook)
	       egg-insert-after-hook
	       (run-hooks 'egg-insert-after-hook))
	  (if overwrite-mode
	      (let ((str (buffer-substring p (point))))
		(delete-text-in-column nil (+ (current-column) (string-width str)))))
	  (if (and (boundp 'self-insert-after-hook)
		   self-insert-after-hook)
	      (funcall self-insert-after-hook
		       p
		       (point)))
	  (if (fboundp 'egg:do-auto-fill)
	      (egg:do-auto-fill)
	    (and auto-fill-function
		 (> (current-column) fill-column)
		 (funcall auto-fill-function))))))
    (and (markerp tcode-bushu-henkan-last-point)
	 (set-marker tcode-bushu-henkan-last-point nil))
    (setq tcode-bushu-henkan-last-point nil)
    )))

(defun tcode-insert-register (reg arg)
  "insert-register $B$HF1$8!#$?$@$7!"%]%$%s%H$H%^!<%/$N0LCV$,5U!#(B"
  (interactive "cInsert register: \nP")
  (insert-register reg (not arg)))

;;
;; Egg patches
;;

(defun tcode-fence-exit-mode ()
  (interactive)
  (fence-exit-mode)
  (tcode-egg-exit))

(defun tcode-henkan-kakutei ()
  (interactive)
  (henkan-kakutei)
  (tcode-egg-exit))

(defun henkan-kakutei-and-self-insert ()
  (interactive)
  (toggle-egg-mode)
  (setq unread-command-events (cons last-command-event unread-command-events))
  (henkan-kakutei)
  (tcode-egg-exit))

(defun tcode-egg-exit ()
  (setq egg:*mode-on* nil
	egg:*in-fence-mode* nil)
  (setq tcode-in-mazegaki-p nil)
  (tcode-mode-line-redisplay))

(if (featurep 'egg)
    (progn
      (substitute-key-definition 'fence-exit-mode 'tcode-fence-exit-mode fence-mode-map)
      (substitute-key-definition 'henkan-kakutei 'tcode-henkan-kakutei henkan-mode-map)))

;;
;; Help
;;
(defun tcode-query-stroke (p)
  "$B0LCV(BP$B$K$"$kJ8;z$NBG$AJ}$rI=<($9$k(B"
  (interactive "d")
  (let ((ch (buffer-substring (1+ p) p)))
    (tcode-display-stroke-for-char ch)))

(defun tcode-display-stroke-for-char (ch)
  "CH($BA43Q0lJ8;z(B)$B$NBG$AJ}$rI=<($9$k(B"
  (interactive "sHelp:")
  (let ((stroke (tcode-stroke-for-char ch))
	(buf (get-buffer-create " *stroke2*"))
	k1 k2)
      (save-excursion
	(set-buffer buf)
	(erase-buffer)
	(if stroke
	    (progn
	      (setq k1 (tcode-get-key-location (car stroke)))
	      (setq k2 (tcode-get-key-location (cdr stroke)))
	      (insert "\
                      
$B!&!&!&!&!!!!!!!&!&!&!&(B
$B!&!&!&!&!!!!!!!&!&!&!&(B
$B!&!&!&!&!!!!!!!&!&!&!&(B")
	      (if (equal k1 k2)
		  (progn
		    (tcode-help-stroke k1 tcode-help-same-stroke)
		    (goto-line 3)
		    (end-of-line)
		    (insert "     " tcode-help-same-stroke ": $BBh0l!"BhFsBG80(B"))
		(tcode-help-stroke k1 tcode-help-first-stroke)
		(tcode-help-stroke k2 tcode-help-second-stroke)
		(goto-line 3)
		(end-of-line)
		(insert "     " tcode-help-first-stroke ": $BBh0lBG80(B")
		(goto-line 4)
		(end-of-line)
		(insert "     " tcode-help-second-stroke ": $BBhFsBG80(B"))))
	(goto-char (point-min))
	(end-of-line)
	(insert "     " ch)
	(let ((decomposed (tcode-decompose-char ch)))
	  (if decomposed
	      (insert " = " (car decomposed) " + " (cdr decomposed))))
	(with-output-to-temp-buffer " *tcode-help*"
	  (princ (buffer-substring (point-min) (point-max)))))))

(defun tcode-help-stroke (loc ch)
  "subroutine of tcode-display-stroke-for-char"
  (goto-line (1+ (car loc)))
  (move-to-column (+ (* 2 (cdr loc)) (if (>= (cdr loc) 6) 0 -2)))
  (delete-char 1)
  (insert ch))

(defun tcode-show-tables (first second)
  (interactive)
  (let* ((buf (draw-tcode-tables first second)))
    (if tcode-auto-zap-table
	(save-window-excursion
	  (with-output-to-temp-buffer " *tcode-help*"
	    (save-excursion
	      (set-buffer buf)
	      (princ (buffer-substring (point-min) (point-max)))))
	  (sit-for 0)
	  (setq unread-command-events (cons (read-char) unread-command-events)))
      (with-output-to-temp-buffer " *tcode-help*"
	(save-excursion
	  (set-buffer buf)
	  (princ (buffer-substring (point-min) (point-max))))))))

(defun tcode-make-table-line (k1 k2)
  "subroutine of tcode-make-LR-block"
  (let ((i1 k1) (j 0) c)
    (while (< j 5)
      (setq c (aref (aref tcode-tbl k2) i1))
      (insert c)
      (if (= (char-width c) 1) (insert " "))
      (setq j (1+ j) i1 (1+ i1)))))

(defun tcode-make-LR-block (k1-start k2-start)
  "subroutine of tcode-insert-stroke-file"
  (let ((k1 k1-start) (k2 k2-start) y x yy)
    (setq yy 0)
    (while (< yy 4)
      (setq y 0)
      (setq k1 k1-start)
      (while (< y 4)
	(setq k2 k2-start)
	(insert "\n  ")
	(setq x 0)
	(while (< x 5)
	  (tcode-make-table-line k1 k2)
	  (insert "  ")
	  (setq x (1+ x))
	  (setq k2 (1+ k2)))
	(setq y (1+ y))
	(setq k1 (+ 10 k1))
	)
      (setq k2-start (+ 10 k2-start))
      (insert "\n")
      (setq yy (1+ yy))
      )))

(defun tcode-insert-stroke-file ()
  "Make the stroke table and writeout the content into
tcode stroke help table file if it is older than tcode
table."
  (erase-buffer)
  (if (and (file-exists-p tcode-stroke-file-name)
	   (file-newer-than-file-p tcode-stroke-file-name tcode-table-file-name))
      (insert-file-contents tcode-stroke-file-name)
    (insert "\nLL\n")
    (tcode-make-LR-block 0 0)
    (insert "\nLR\n")
    (tcode-make-LR-block 0 5)
    (insert "\nRL\n")
    (tcode-make-LR-block 5 0)
    (insert "\nRR\n")
    (tcode-make-LR-block 5 5)
    (write-file tcode-stroke-file-name)))

(defun draw-tcode-tables (first second)
  "Draw tcode stroke tables.
FIRST corresponds to the first stroke. If nil, then left, else right.
SECOND to the second stroke.  If nil, then left, else right."
  (interactive)
  (let ((buf (get-buffer-create " *stroke*"))
	(str (concat "^" (if first "R" "L") (if second "R" "L"))))
    (save-excursion
      (set-buffer buf)
      (widen)
      (goto-char (point-min))
      (if (re-search-forward str nil  t)
	  (progn
	    (forward-line 0)
	    (let ((top (point))
		  (end
		   (progn
		     (forward-line 3)
		     (if (re-search-forward "^[RL]" nil t)
			 (progn(forward-line -1)(point))
		       (point-max)))))
	      (narrow-to-region top end)
	      buf))
	(tcode-insert-stroke-file)
	(draw-tcode-tables first second)))))

;;
;; JIS$B%3!<%IF~NO(B
;;

(defvar tcode-jiscode-buffer " *jis-code*")
(defvar tcode-jiscode-last-selected-buffer nil)

(defvar tcode-jiscode-map nil)
(if tcode-jiscode-map ()
  (setq tcode-jiscode-map (make-sparse-keymap))
  (define-key tcode-jiscode-map " " 'tcode-jiscode-insert)
  (define-key tcode-jiscode-map "\C-c" 'exit-recursive-edit))

(defun tcode-jiscode-insert-line (1st 2nd)
  (insert (format "%02x%02x " (- 1st 128) (- 2nd 128)))
  (let ((i 0))
    (while (and (< i 16) (< 2nd 255))
      (insert (make-character lc-jp 1st 2nd))
      (setq 2nd (1+ 2nd))
      (setq i (1+ i)))
    (insert " : ")
    (while (and (< i 32)(< 2nd 255))
      (insert (make-character lc-jp 1st 2nd))
      (setq 2nd (1+ 2nd))
      (setq i (1+ i))))
  (insert ":\n"))

(defun tcode-jiscode-insert-tables ()
  "JIS$B%3!<%II=$r:n$C$F(B, $B8=%P%C%U%!$KA^F~$9$k(B. "
  (message "Making jiscode tables...")
  (let ((1st 161) (2nd 161))
    (while (< 1st 255)
      (setq 2nd 161)
      (tcode-jiscode-insert-line 1st 2nd)
      (tcode-jiscode-insert-line 1st (+ 2nd 32))
      (tcode-jiscode-insert-line 1st (+ 2nd (* 2 32)))
      (setq 1st (1+ 1st)))
    )
  (message "Making jiscode tables...done.")
  (beginning-of-buffer))

(defun tcode-jiscode-insert (arg)
  (interactive "p")
  (save-excursion
    (let* ((p (point))
	   (str (buffer-substring p (1+ p))))
      (set-buffer tcode-jiscode-last-selected-buffer)
      (tcode-insert str))))

(defun tcode-start-jiscode ()
  "JIS$B4A;zI=$r%P%C%U%!(B*jis-code*$B$KI=<($7(B, $BBPOCE*$K4A;z$rA*$s$G85$N%P%C%U%!$KA^F~$G$-$k$h$&$K$9$k(B. 
  *jis-code*$BCf$G;H$($k%-!<$O<!$NDL$j(B:

   Space :	$B%+!<%=%k0LCV$N4A;z$rA0$$$?%P%C%U%!$KA^F~$9$k(B. 
   C-c   :	$BA0$$$?%P%C%U%!$K$b$I$k(B. "  
  (interactive)
  (save-window-excursion
    (setq tcode-jiscode-last-selected-buffer (current-buffer))
    (switch-to-buffer-other-window tcode-jiscode-buffer)
    (if (= (point-min) (point-max))
	(progn
	  (tcode-jiscode-insert-tables)
	  (setq mode-name "JIS-select")
	  (use-local-map tcode-jiscode-map)
	  (setq major-mode 'tcode-start-jiscode)
	  (setq buffer-read-only t)))
    (message "SPC to pick the character; C-c to exit")
    (recursive-edit)))

;;
;; $B8r$<=q$-<-=q4IM}(B
;;

(defun tcode-kill-emacs-hook ()
  (tcode-save-jisyo)
  (tcode-record))

(add-hook 'kill-emacs-hook 'tcode-kill-emacs-hook)

(defun tcode-save-jisyo ()
  (interactive)
  (if (not (memq mazegaki-buffer (buffer-list))) nil ; buffer is deleted?
    (if (not (buffer-modified-p mazegaki-buffer)) nil ; buffer is not modified
      (let ((filename (expand-file-name tcode-dictionary-name)))
	(save-window-excursion
	  (rename-file filename (concat filename ".BAK") t)
	  (set-buffer mazegaki-buffer)
	  (write-file filename))))))

(defun tcode-record ()
  (and tcode-record-file-name
       (> tcode-number-strokes 0)
       (save-excursion
	 (set-buffer (get-buffer-create " *record*"))
	 (erase-buffer)
	 (insert
	  (format "%s  $BJ8;z(B: %4d  $BIt<s(B: %3d(%d%%)  $B8r$<=q$-(B: %3d(%d%%)  $B5!G=(B: %3d(%d%%)\n"
		  (let ((time (current-time-string)))
		    (if (not (string-match "^... \\(.*:.*\\):" time))
			""
		      (substring time (match-beginning 1) (match-end 1))))
		  tcode-input-chars
		  tcode-bushu-occurrence
		  (/ (* 100 tcode-bushu-occurrence) tcode-number-strokes)
		  tcode-mazegaki-occurrence
		  (/ (* 100 tcode-mazegaki-occurrence) tcode-number-strokes)
		  tcode-special-occurrence
		  (/ (* 100 tcode-special-occurrence) tcode-number-strokes)))
	 (append-to-file (point-min) (point-max) tcode-record-file-name))))

(defun tcode-find-entry (yomi)
  (widen)
  (goto-char(point-min))
  (let ((yomi-s (concat "^" yomi " ")))
    (if (re-search-forward yomi-s nil t)
	(point)
      nil)))

(defun tcode-toroku (yomi kanji)
  "$BFI$_$r(BYOMI$B!"4A;z$r(BKANJI$B$H$7$F8r$<=q$-<-=q$KEPO?$9$k(B"
  (interactive "s$BFI$_(B:\ns$B4A;z(B:")
  (save-excursion
    (set-buffer mazegaki-buffer)
    (if (tcode-find-entry yomi)
	(progn
	  (let ((kanji-s (concat "/" kanji "/")))
	    (if (search-forward kanji-s
				(save-excursion (end-of-line)(point)) t)
		(message "$BC18l$O$9$G$KEPO?$5$l$F$$$^$9(B")
	      (end-of-line)
	      (insert kanji "/"))))
      (insert yomi " " "/" kanji "/\n"))
    ))
(defun tcode-toroku-region (start end)
  (interactive "r")
  (let ((yomi (read-string "$BFI$_(B:")))
    (tcode-toroku yomi (buffer-substring start end))))

(defun tcode-purge-from-jisho (yomi kanji)
  "$BFI$_(BYOMI$B!"4A;z(BKANJI$B$N%(%s%H%j$r<-=q$+$i:o=|$9$k(B"
  (interactive "s$BFI$_(B:\ns$B4A;z(B:")
  (save-excursion
    (set-buffer mazegaki-buffer)
    (if (tcode-find-entry yomi)
	(let ((kanji-s (concat "/" kanji "/"))
	      (kanji-j (concat " /" kanji "/$")))
	  (beginning-of-line)
	  (cond
	   ((re-search-forward kanji-j
			       (save-excursion (end-of-line)(point)) t)
	    (beginning-of-line)
	    (let ((p (point)))
	      (delete-region p (progn(end-of-line)(1+(point))))))
	   ((search-forward kanji-s
			    (save-excursion (end-of-line)(point)) t)
	    (delete-region (- (point) (length kanji) ) (point)))
	   (t
	    (error "$BFI$_(B[%s]$B!"4A;z(B[%s]$B$NC18l$OEPO?$5$l$F$$$^$;$s(B"
		   yomi kanji)
	    )))
      (error "$BFI$_(B[%s]$B!"4A;z(B[%s]$B$NC18l$OEPO?$5$l$F$$$^$;$s(B"
	     yomi kanji)
      )))

(defun tcode-gakushu (yomi kanji)
  "YOMI,KANJI$B$G<-=q$r3X=,$5$;$k!#$3$N4X?t$r8F$VA0$K!"<-=q$N%+!<%=%k0LCV$O(B
YOMI$B$N$"$k9T$K$J$1$l$P$$$1$J$$!#(B"
  (if (> tcode-mazegaki-gakushu-level 0)
      (save-excursion
	(set-buffer mazegaki-buffer)
	(widen)
	(beginning-of-line)
	(let ((yomi-s (concat "/" kanji "/")))
	  (if (search-forward yomi-s (save-excursion(end-of-line)(point)) t)
	      (progn
		(delete-region (- (point) (length kanji)) (point))
		(beginning-of-line)
		(search-forward "/")
		(insert kanji "/"))))
	(if (> tcode-mazegaki-gakushu-level 1)
	    (progn
	      (beginning-of-line)
	      (let* ((beg (point))
		     (end (save-excursion (end-of-line) (1+ (point))))
		     (yy (buffer-substring beg end)))
		(delete-region beg end)
		(goto-char (point-min))
		(insert yy)))))))

(defun tcode-mazegaki-get-applicants (s)
  "S$B$r$h$_$H$9$k=O8l$N%j%9%H$rJV$9(B"
  (tcode-mazegaki-get-applicants-s s))

(defun tcode-mazegaki-get-applicants-s (s)
  (let (applicants
	(yomi-s (concat "^" s " ")))
    (save-excursion
      (set-buffer mazegaki-buffer)
      (widen)
      (goto-char (point-min))
      (if (re-search-forward yomi-s nil t)
	  (let (p)
	    (narrow-to-region (1+ (point)) (save-excursion (end-of-line)(point)))
	    (goto-char (point-min))
	    (setq p (point))
	    (while (re-search-forward "/" nil t)
	      (setq applicants (cons (buffer-substring p (1- (point))) applicants))
	      (setq p (1+ (point)))))
	)
      (nreverse applicants)
      )))


;;
;; $B8r$<=q$-JQ49(B
;;

;; Many parts are stolen from skk.el

(defun tcode-search-for-marker (s)
  (save-excursion
    (if (search-backward s nil t) (progn (forward-char 1) (point))
      nil)))

(defvar tcode-old-map nil)
(defvar tcode-mazegaki-input-map nil
  "*$B8r$<=q$-JQ49$NJ8;zF~NOCf$N%-!<%^%C%W(B")
(defvar tcode-mazegaki-sentaku-map nil
  "*$B8r$<=q$-JQ49$N8uJdA*BrCf$N%-!<%^%C%W(B")
(defvar tcode-current-kouho 0)
(defvar tcode-kouho-max 0)
(defvar tcode-mazegaki-applicants nil)
(defvar tcode-mazegaki-yomi nil)

(setq tcode-mazegaki-input-map (make-sparse-keymap))
(setq tcode-mazegaki-sentaku-map (make-sparse-keymap))
(let ((i 0) map)
  (while (< i 126)
    (setq map (cons (cons i 'tcode-mazegaki-sentaku-insert) map))
    (setq i (1+ i))
    )
  (setq tcode-mazegaki-sentaku-map (cons 'keymap map)))

(define-key tcode-mazegaki-input-map " " 'tcode-mazegaki-sentaku-start)
(define-key tcode-mazegaki-input-map "\C-k" 'tcode-clear)
(define-key tcode-mazegaki-input-map "\C-t" 'tcode-mazegaki-sentaku-toroku)

(define-key tcode-mazegaki-sentaku-map "\C-k" 'tcode-mazegaki-kakutei-gakushu)
(define-key tcode-mazegaki-sentaku-map "\C-n" 'tcode-mazegaki-next-applicant)
(define-key tcode-mazegaki-sentaku-map " " 'tcode-mazegaki-next-applicant)
(define-key tcode-mazegaki-sentaku-map "\C-i" 'tcode-mazegaki-unhenkan)
(define-key tcode-mazegaki-sentaku-map "\C-p" 'tcode-mazegaki-previous-applicant)
(define-key tcode-mazegaki-sentaku-map "\C-m" 'tcode-mazegaki-kakutei-gakushu)
(define-key tcode-mazegaki-sentaku-map "\C-x" 'tcode-mazegaki-delete)

(defun tcode-mazegaki-delete ()
  (interactive)
  (if (yes-or-no-p
       (concat "$BFI$_(B[" tcode-mazegaki-yomi
	       "] $B4A;z(B["
	       (nth tcode-current-kouho tcode-mazegaki-applicants)
	       "] $B$r<-=q$+$i:o=|$7$^$9(B "))
      (tcode-purge-from-jisho tcode-mazegaki-yomi
			      (nth tcode-current-kouho tcode-mazegaki-applicants))))

(defun tcode-start-mazegaki ()
  (save-excursion
    (setq mazegaki-buffer (get-buffer-create " *mazegaki-dictionary*"))
    (set-buffer mazegaki-buffer)
    (widen)
    (if (= (point-min) (point-max))
	(progn
	  (erase-buffer)
	  (insert-file-contents tcode-dictionary-name)
	  (set-buffer-modified-p nil))))
  (if tcode-in-mazegaki-p
      (tcode-mazegaki-kakutei)
    (insert "$B"$(B")
    (progn
      (setq tcode-old-map (current-local-map))
      (use-local-map tcode-mazegaki-input-map))
    (setq tcode-in-mazegaki-p t)))

(defun tcode-mazegaki-replace-kouho (str)
  "$B8r$<=q$-8uJd$r(BSTR$B$KF~$l$+$($k(B"
  (let* ((cur (point))
	 (first (tcode-search-for-marker "$B"$(B"))
	 (yomi (buffer-substring first cur)))
    (delete-region first cur)		;delete yomi
    (insert str)))

(defun tcode-mazegaki-sentaku-start ()
  (interactive)
  (let* ((cur (point))
	 (first (tcode-search-for-marker "$B"$(B"))
	 mazegaki-end
	 pp)
    (setq tcode-mazegaki-yomi (buffer-substring first cur))
    (setq tcode-mazegaki-applicants
	  (tcode-mazegaki-get-applicants tcode-mazegaki-yomi))
    (setq tcode-current-kouho 0)
    (setq tcode-kouho-max (length tcode-mazegaki-applicants))
    (if (null tcode-mazegaki-applicants)
	(progn
	  (beep)
	  (message "$B8+$D$+$j$^$;$s!#(B"))
      (use-local-map tcode-mazegaki-sentaku-map)
      (goto-char first)
      (delete-region first cur)
      (insert (car tcode-mazegaki-applicants))
      )))

(defun tcode-mazegaki-unhenkan ()
  "$B4A;z$NA*Br%b!<%I$KF~$C$F$$$k$N$r$L$1$F!"(B
$BFI$_$NF~NO>lLL$K$b$I$9(B"
  (interactive)
  (use-local-map tcode-mazegaki-input-map)
  (tcode-mazegaki-replace-kouho tcode-mazegaki-yomi))

(defun tcode-mazegaki-sentaku-toroku ()
  "$BFI$_$NF~NO>lLL$G!"$=$NFI$_$KBP1~$9$k(B
$B=O8l$rEPO?$9$k!#(B"
  (interactive)
  (let* ((cur (point))
	 (kanji (read-string "$BEPO?(B:" nil)))
    (tcode-toroku tcode-mazegaki-yomi kanji)
    (tcode-mazegaki-replace-kouho kanji)
    (tcode-mazegaki-kakutei)
    ))

(defun tcode-mazegaki-next-applicant ()
  (interactive)
  (setq tcode-current-kouho (1+ tcode-current-kouho))
  (if (>= tcode-current-kouho tcode-kouho-max)
      (setq tcode-current-kouho 0))
  (tcode-mazegaki-replace-kouho
   (nth tcode-current-kouho tcode-mazegaki-applicants)))

(defun tcode-mazegaki-previous-applicant ()
  (interactive)
  (setq tcode-current-kouho (1- tcode-current-kouho))
  (if (< tcode-current-kouho 0)
      (setq tcode-current-kouho (1- tcode-kouho-max)))
  (tcode-mazegaki-replace-kouho
   (nth tcode-current-kouho tcode-mazegaki-applicants)))

(defun tcode-mazegaki-sentaku-insert ()
  (interactive)
  (tcode-gakushu tcode-mazegaki-yomi (nth tcode-current-kouho
					  tcode-mazegaki-applicants))
  (setq unread-command-events
	(cons last-command-event unread-command-events))
  (tcode-mazegaki-kakutei))

(defun tcode-mazegaki-kakutei ()
  "$B8r$<=q$-JQ49$r3NDj$7$F!"8r$<=q$-JQ49%b!<%I$rH4$1$k!#(B"
  (interactive)
  (let ((line-beginning
	 (save-excursion (beginning-of-line)
			 (point)))
	(kakutei (nth tcode-current-kouho tcode-mazegaki-applicants))
	p)
    (and tcode-kakutei-register
	 (set-register tcode-kakutei-register kakutei))
    (save-excursion
      (goto-char (tcode-search-for-marker "$B"$(B"))
      (delete-char -1)
      (setq p (point)))
    (setq tcode-in-mazegaki-p nil)
    (if kakutei
	(progn
	  (setq tcode-mazegaki-occurrence
		(+ (chars-in-string kakutei) tcode-mazegaki-occurrence))
	  (if overwrite-mode
	      (delete-text-in-column nil (+ (current-column) (string-width kakutei))))
	  (and (boundp 'self-insert-after-hook)
	       self-insert-after-hook
	       (funcall self-insert-after-hook p (point)))))
    (use-local-map tcode-old-map)
    ))

(defun tcode-mazegaki-kakutei-gakushu ()
  (interactive)
  (tcode-gakushu tcode-mazegaki-yomi (nth tcode-current-kouho
					  tcode-mazegaki-applicants))
  (tcode-mazegaki-kakutei))

;;
;; $BIt<sJQ49(B
;;

(defun tcode-bushu-init (level)
  "$BIt<sJQ49$N=i4|2=$r9T$J$&!#(B
$B0z?t(BLEVEL$B$,JQ?t(Btcode-bushu-on-demand$B$h$j>.$5$+$C$?$i9T$J$o$J$$(B"
  (interactive (list 999))
  (or tcode-bushu-hash
      (< level tcode-bushu-on-demand)
      (progn
	(or (and (boundp 'tcode-tbl) tcode-tbl)
	    (tcode-load-table tcode-table-file-name))
	(tcode-read-bushu-dictionary))))

(defun tcode-read-bushu-dictionary ()
  "$BIt<sJQ49<-=q$rFI$_9~$`(B"
  (interactive)
  (save-excursion
    (message "Hashing bushu dictionary...")
    (setq bushu-buffer (get-buffer-create "*bushu-dictionary*"))
    (setq tcode-bushu-hash (make-vector 1009 0))
    (setq tcode-bushu-reverse-hash (make-vector 1009 0))
    (set-buffer bushu-buffer)
    (erase-buffer)
    (insert-file-contents tcode-bushu-dictionary-name)
    (set-buffer-modified-p nil)
    (while (not (eobp))
      (let (character component)
	(setq component (buffer-substring (point) (progn (forward-char 2) (point))))
	(setq character (buffer-substring (point) (progn (forward-char 1) (point))))
	(set (intern component tcode-bushu-hash) character)
	(set (intern character tcode-bushu-reverse-hash) component)
	(forward-line 1)))
    (kill-buffer bushu-buffer)
    (let ((second 0) first row char)
      (while (< second 40)
	(setq row (aref tcode-tbl second)
	      first 0)
	(while (< first 40)
	  (setq char (intern (char-to-string (aref row first)) tcode-bushu-reverse-hash))
	  (put char 'stroke (cons first second))
	  (setq first (1+ first)))
	(setq second (1+ second))))
    (message "Hashing bushu dictionary...done.")
    (run-hooks 'tcode-bushu-ready-hook)))

(defun tcode-start-bushu ()
  "$BIt<sJQ49$r;O$a$k(B"
  (tcode-bushu-init 2)
  (or tcode-bushu-hash (error "Bushu dictionary not ready."))
  (setq tcode-bushu-nest (1+ tcode-bushu-nest))
  (insert "$B"%(B"))

(defun tcode-end-bushu ()
  "$B0lHVFbB&$NIt<sJQ49$r=*N;$9$k(B"
  (setq tcode-bushu-nest (1- tcode-bushu-nest))
  (let ((line-beginning
	 (save-excursion (beginning-of-line)
			 (point))))
    (save-excursion
      (if (search-backward "$B"%(B" line-beginning t)
	  (delete-char 1)))))

(defun tcode-decompose-char (kanji)
  "KANJI$B$r(B2$B$D$NIt<s$KJ,2r$9$k!#(B
$BJV$jCM$O(B2$B$D$N;z$r(B cons $B$7$?%;%k!#(B
$BJ,2r$G$-$J$$;~$K$O(Bnil$B$rJV$9(B"
  (tcode-bushu-init 2)
  (let ((sym (and tcode-bushu-reverse-hash
		  (intern-soft kanji tcode-bushu-reverse-hash)))
	l)
    (if (and sym (boundp sym))
	(progn
	  (setq sym (symbol-value sym))
	  (setq l (string-to-char-list sym))
	  (cons (char-to-string (car l))
		(char-to-string (car (cdr l)))))
      nil)))

(defconst tcode-lookup-outset-normalization
  '(("$B!J(B" . "(") ("$B!K(B" . ")")
    ("," . "$B!"(B") ("." . "$B!#(B")
    ("$B!$(B" . "$B!"(B") ("$B!%(B" . "$B!#(B")
    ("$B#0(B" . "0") ("$B#1(B" . "1") ("$B#2(B" . "2") ("$B#3(B" . "3")
    ("$B#4(B" . "4") ("$B#5(B" . "5") ("$B#6(B" . "6") ("$B#7(B" . "7")
    ("$B#8(B" . "8") ("$B#9(B" . "9")
    ("$B!s(B" . "%") ("$B!)(B" . "?") ("$B!*(B" . "!")))

(defun tcode-lookup-outset-sub (tchar-1 tchar-2)
  "TCHAR-1 $B$H(B TCHAR-2 $B$rD>@\AH$_9g$o$;$F$G$-$k30;z$rA\$9!#(B
$B8+$D$+$i$J$+$C$?>l9g$K$O(B nil $B$rJV$9(B"
  (setq tchar-1
	(or (cdr (assoc tchar-1 tcode-lookup-outset-normalization))
	    tchar-1))
  (setq tchar-2
	(or (cdr (assoc tchar-2 tcode-lookup-outset-normalization))
	    tchar-2))
  (let ((sym (intern-soft (concat tchar-1 tchar-2) tcode-bushu-hash)))
    (or (and sym (eval sym))
	(progn
	  (setq sym (intern-soft (concat tchar-2 tchar-1) tcode-bushu-hash))
	  (and sym (eval sym))))))

(defun tcode-outset-alternative (char)
  "CHAR $B$HEy2A$JJ8;z(B EQCHAR $B$rJV$9!#FC$K$J$1$l$P(B CHAR $B<+?H$rJV$9!#(B
CHAR $B$,Ey2A$G$"$k$H$O(B NL + EQCHAR = CHAR $B$HDj5A$5$l$F$$$k$3$H$G$"$k!#(B"
  (let ((eqchar (tcode-decompose-char char)))
    (if (and eqchar (string= (car eqchar) 'N))
	(cdr eqchar)
      char)))

(defun tcode-lookup-outset (tchar-1 tchar-2)
  "TCHAR-1 $B$H(B TCHAR-2 $B$rAH$_9g$o$;$F$G$-$k30;z$rA\$9!#(B
$B8+$D$+$i$J$+$C$?>l9g$K$O(B nil $B$rJV$9(B"
  (setq tchar-1 (tcode-2-to-1 tchar-1)
	tchar-2 (tcode-2-to-1 tchar-2))
  (or (tcode-lookup-outset-sub tchar-1 tchar-2)
      (progn
	(setq tchar-1 (tcode-outset-alternative tchar-1))
	(setq tchar-2 (tcode-outset-alternative tchar-2))
	(or (tcode-lookup-outset-sub tchar-1 tchar-2)
	    (let (tc11 tc12 tc21 tc22 loop i (res nil))
	      (if (setq tc11 (tcode-decompose-char tchar-1))
		  (progn (setq tc12 (cdr tc11))
			 (setq tc11 (car tc11))))
	      (if (setq tc21 (tcode-decompose-char tchar-2))
		  (progn (setq tc22 (cdr tc21))
			 (setq tc21 (car tc21))))
	      (and tc11 (string= tc11 'N) (setq tc11 nil))
	      (and tc12 (string= tc12 'N) (setq tc12 nil))
	      (and tc21 (string= tc21 'N) (setq tc21 nil))
	      (and tc22 (string= tc22 'N) (setq tc22 nil))
	      (or (and tc11 (string= tc11 tchar-2)
		       (not (string= tc12 tchar-1))
		       (not (string= tc12 tchar-2))
		       tc12); subtraction
		  (and tc12 (string= tc12 tchar-2)
		       (not (string= tc11 tchar-1))
		       (not (string= tc11 tchar-2))
		       tc11)
		  (and tc21 (string= tc21 tchar-1)
		       (not (string= tc22 tchar-1))
		       (not (string= tc22 tchar-2))
		       tc22)
		  (and tc22 (string= tc22 tchar-1)
		       (not (string= tc21 tchar-1))
		       (not (string= tc21 tchar-2))
		       tc21)
		  (progn
		    (setq loop '((tchar-1 . tc22) (tchar-2 . tc12) (tchar-1 . tc21)
				 (tchar-2 . tc11) (tc12 . tc22) (tc21 . tc12)
				 (tc11 . tc22) (tc21 . tc11)))
		    (while (and loop (not res))
		      (setq i (car loop))
		      (setq loop (cdr loop))
		      (let ((i1 (eval (car i))) (i2 (eval (cdr i))))
			(setq res (and i1 i2 (tcode-lookup-outset-sub i1 i2)))
			(if (or (not res) (string= res tchar-1) (string= res tchar-2))
			    (setq res nil))))
		    res)
		  ;; new subtraction
		  (and tc11 tc21 (string= tc11 tc21)
		       (not (string= tc12 tchar-1))
		       (not (string= tc12 tchar-2)) tc12)
		  (and tc11 tc22 (string= tc11 tc22)
		       (not (string= tc12 tchar-1))
		       (not (string= tc12 tchar-2)) tc12)
		  (and tc12 tc21 (string= tc11 tc21)
		       (not (string= tc11 tchar-1))
		       (not (string= tc11 tchar-2)) tc11)
		  (and tc12 tc22 (string= tc11 tc21)
		       (not (string= tc11 tchar-1))
		       (not (string= tc11 tchar-2)) tc11)))))))

(defun tcode-get-prev-nonspace ()
  "Return preceding non-space letter.  Move point to the beg' of that char."
  (while (eq (preceding-char) ? )
    (forward-char -1))
  (prog1 (char-to-string (or (preceding-char) 0))
    (forward-char -1)))

(defun tcode-compose-chars ()
  "$B%]%$%s%H$NA0$NFsJ8;z$r9g@.$9$k(B"
  (interactive "*")
  (tcode-bushu-init 3)
  (let* ((p (point))
	 (prev-char (tcode-get-prev-nonspace))
	 (p1 (point))
	 (prev-prev-char (tcode-get-prev-nonspace))
	 (p2 (point))
	 (kanji (tcode-lookup-outset prev-prev-char prev-char)))
    (goto-char p)
    (if kanji
	(progn
	  (setq tcode-bushu-occurrence (1+ tcode-bushu-occurrence))
	  (delete-region p2 p)
	  (insert kanji)
	  (setq tcode-help-char kanji))
      (beep)
      (message ""))))

;;
;; incremental search
;;

(load "tc-isearch19")

;;
;; Tcode$B$N%-!<%F!<%V%kI=(B
;;

(defun tcode-load-table (filename)
  (interactive "fLoad T-Code table file: ")
  (load filename)
  (let (k1 newval)
    (setq k1 0)
    (while (< k1 40)
      (setq newval (vconcat (delq ?  (string-to-char-list (aref tcode-tbl k1)))))
      (or (= (length newval) 40)
	  (error "Table corrupted at line %d." (1+ k1)))
      (aset tcode-tbl k1 newval)
      (setq k1 (1+ k1)))))

(defun tcode-encode-key (k1 k2 c1 c2)
  "Encode the 1st Tcode stroke(K1) and 2nd stroke(K2) to a Kanji character."
  (if (or (< k1 0) (< k2 0))
      (concat (char-to-string c1)
	      (if (= c2 ?\ ) ""
		(char-to-string c2)))
    (let* ((ch (char-to-string (aref (aref tcode-tbl k2) k1)))
	   (tcode-kutouten
	    (cond ((null tcode-kutouten) '(("$B!"(B" . ", ") ("$B!#(B" . ". ")))
		  ((eq t tcode-kutouten) nil)
		  (t tcode-kutouten)))
	   (new (assoc ch tcode-kutouten)))
      (if (and new
	       (= 0 tcode-bushu-nest))
	  (cdr new)
	ch))))

(defun tcode-stroke-for-char (ch)
  "CH$B$NBG$AJ}$r(B(first . second)$B$GJV$9!#D>@\F~NO$G$-$J$1$l$P(Bnil$B$rJV$9!#(B"
  (if tcode-bushu-reverse-hash
      (get (intern-soft ch tcode-bushu-reverse-hash) 'stroke)
    (let (first row (second 0))
      (catch 'found
	(if (stringp ch) (setq ch (string-to-char ch)))
	(while (< second 40)
	  (setq row (aref tcode-tbl second)
		first 0)
	  (while (< first 40)
	    (if (= (aref row first) ch)
		(throw 'found (cons first second)))
	    (setq first (1+ first)))
	  (setq second (1+ second)))
	nil))))

(defun tcode-transpose-strokes (arg)
  "$B%]%$%s%H0LCV$NJ8;z$N%9%H%m!<%/$rF~$l$+$($k(B"
  (interactive "*P")
  (if (not (tcode-on))
      (transpose-chars arg)
    (if (eolp) (forward-char -1))
    (let ((ch (buffer-substring (point) (save-excursion (forward-char 1) (point))))
	  stroke)
      (setq stroke (tcode-stroke-for-char ch))
      (and stroke
	   (setq ch (tcode-encode-key (cdr stroke) (car stroke) ?1 ?1 ))
	   (progn
	     (delete-char 1)
	     (insert ch))))))

(defun tcode-get-key-location (address)
  (cons (/ address 10) (1+ (% address 10))))

(defun tcode-key-address-right-p (address)
  (let ((location (tcode-get-key-location address)))
    (>= (cdr location) 6)))
(defun tcode-key-address-left-p (address)
  (let ((location (tcode-get-key-location address)))
    (< (cdr location) 6)))
(defun tcode-unmap-key (c)
  (let ((max (length tcode-keymap-table))
	(i 0)
	(looping t))
    (while (and (< i max) looping)
      (if (= c (aref tcode-keymap-table i))
	  (progn
	    (setq looping nil))
	(setq i (+ i 1))))
    (+ i 32)))

(defun tcode-get-key-address (c)
  (if (or (< c ? ) (> c ?~)) -1
    (aref tcode-keymap-table (- c 32))))



;;
;; 2-byte $B1Q?t;z(B
;;
(defvar tcode-alnum-2byte-regexp nil)

(defun tcode-check-alnum-1-to-2-table ()
  (if (stringp tcode-alnum-1-to-2-table)
      (setq tcode-alnum-1-to-2-table
	    (vconcat (string-to-char-list tcode-alnum-1-to-2-table))
	    tcode-alnum-2byte-regexp nil))
  (if (null tcode-alnum-2byte-regexp)
      (setq tcode-alnum-2byte-regexp
	    (concat "["
		    (mapconcat 'char-to-string tcode-alnum-1-to-2-table "")
		    "]+"))))

(defun tcode-1-to-2-region (beg end)
  "$B%j!<%8%g%sFb$N(B1$B%P%$%H1Q?t;z$r(B2$B%P%$%H$KJQ49$9$k(B"
  (interactive "*r")
  (tcode-check-alnum-1-to-2-table)
  (let (char)
  (save-excursion
    (save-restriction
      (goto-char beg)
      (narrow-to-region beg end)
      (while (progn (skip-chars-forward "^!-~" (point-max))
		    (< (point) (point-max)))
	(setq char (following-char))
	(delete-char 1)
	(insert (tcode-1-to-2 char)))))))

(defun tcode-1-to-2 (str)
  "STR $BFb$N(B1$B%P%$%H1Q?t;z$r(B2$B%P%$%H$KJQ49$9$k!#(B
STR $B$,J8;z$G$"$C$?$i$=$l$r(B2$B%P%$%H$KJQ49$7$?J8;z$rJV$9!#(B"
  (tcode-check-alnum-1-to-2-table)
  (cond ((integerp str)
	 (if (and (<= ?! str) (<= str ?~))
	     (char-to-string (aref tcode-alnum-1-to-2-table (- str ? )))
	   (char-to-string str)))
	((= 1 (length str))
	 (tcode-1-to-2 (aref str 0)))
	(t
	 (mapconcat 'tcode-1-to-2 (string-to-char-list str) ""))))

(defun tcode-2-to-1-region (beg end)
  "$B%j!<%8%g%sFb$N(B2$B%P%$%H1Q?t;z$r(B1$B%P%$%H$KJQ49$9$k(B"
  (interactive "*r")
  (tcode-check-alnum-1-to-2-table)
  (save-excursion
    (save-restriction
      (let (str)
	(goto-char beg)
	(narrow-to-region beg end)
	(while (re-search-forward tcode-alnum-2byte-regexp nil t)
	  (setq str (buffer-substring (match-beginning 0) (match-end 0)))
	  (delete-region (match-beginning 0) (match-end 0))
	  (insert (tcode-2-to-1 str)))))))

(defun tcode-2-to-1 (str)
  "STR $BFb$N(B2$B%P%$%H1Q?t;z$r(B1$B%P%$%H$KJQ49$9$k(B"
  (tcode-check-alnum-1-to-2-table)
  (cond ((integerp str)
	 (char-to-string
	  (let ((ch 0))
	    (catch 'found
	      (while (< ch 95)
		(if (= (aref tcode-alnum-1-to-2-table ch) str)
		    (throw 'found (+ ch 32)))
		(setq ch (1+ ch)))
	      str))))
	(t
	 (mapconcat 'tcode-2-to-1 (string-to-char-list str) ""))))

(tcode-bushu-init 0)
